/* -*- coding: utf-8 -*-
 * TML
 *
 * Copyright (C) 2010, 2014 by Henry Kroll III, www.thenerdshow.com
 * 
 * This program is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the LGPL in a file named
 * COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */
%option noyywrap
%option stack
%option never-interactive
%s body p bq h bo it table div xlist sup sub strike
%x head title meta keywords style icon script code span inc but tag

%top{
#include "tml.h"
#define YY_BUF_SIZE 65535
}

%%
    static int header_sent,level=0,indent_level=0,indents=0;
    static int caller;
    char *xurl, *tmp;
    (void)yy_top_state; /* not used */
    if(raw){
        header_sent=1;
        BEGIN(body);
    }else BEGIN(head);

<head>^[^<\n]*"title:"[\t ]+ {caller=YY_START;
        BEGIN(title);
    }
<title>.*$ {
    printf("<title>%s</title>",yytext);
    printf("\n<meta name=\"generator\" content=\"Text Markup Language (TML) http://anch.org/tml.html\" />");
    BEGIN(caller);
}

<head>^[^<\n]*("author"|"keywords"|"description")":"[\t ]+ {caller=YY_START;
        xurl=strchr(yytext,':');*xurl='\0';
        strcpy(buf,yytext);
        BEGIN(meta);
    }
<meta>.*$ {
    printf("<meta name=\"%s\" content=\"%s\" />",buf,yytext);
    BEGIN(caller);
}

<head>^[^<\n]*"style:"[\t ]+ {caller=YY_START;
        BEGIN(style);level++;
    }
<style>.*$ {
    if(level==1) printf("<link rel=\"stylesheet\" type=\"text/css\" href=\"%s\" title=\"preferred\" />",yytext);
    else printf("<link rel=\"alternate stylesheet\" type=\"text/css\" href=\"%s\" title=\"alternate%i\" />",yytext,level-1);
    BEGIN(caller);
}

<head>^[^<\n]*"icon:"[\t ]+ {caller=YY_START;
        BEGIN(icon);
    }
<icon>.*$ {
    printf("<link rel=\"icon\" href=\"%s\">",yytext);
    //~ printf("<link rel=\"icon\" type=\"image/vnd.microsoft.icon\" href=\"%s\" />",yytext);
    BEGIN(caller);
}

<head>^[^<\n]*"script:"[\t ]+ {caller=YY_START;
        BEGIN(script);
    }
<script>.*$ {
    printf("<script type=\"text/javascript\" src=\"%s\"></script>",yytext);
    BEGIN(caller);
}

    /* horizontal rule <hr> */
<head,body>^"-"{3,}.*$   {
    /* the first horizontal rule separates the head from the document body */
    if (!header_sent){ /* begin document body */
        puts("</head>\n<body>");
        header_sent = 1;
        BEGIN(body);
    } else puts("\n\n<!-- --> <hr> <!-- -->");
}

    /* =Heading 1= */
    /* ==Heading 2== */
<body>^=+(#[[:alnum:]_\-.]+)? {
    if(!strchr(yytext,'#')){
        level=yyleng;printf("<h%i>\n\n",level);
    } else {
        level=strspn(yytext,"=");
        xurl=yytext+level+1;
        printf("<h%i id=\"%s\">\n\n",level,xurl);
        yyless(level+1); /* put back all but # */
    }
    yy_push_state(h);
}
<h>{
    "="+$  {
        yy_pop_state();printf("</h%i>",level);
    }
}

    /* paragraph */
<body>[\n]{0,2}[^\n\t <{=-]([^\|\n]+\n?)+[^=\n\|>]\n\n {
    char *i;
    strncpy(buf,yytext,7); buf[7]='\0';
    if(strstr(buf,"class")||strstr(buf,"style")||strstr(buf,"id="))REJECT;
    if(!strstr(buf,"[#")&&(level=strcspn(buf,"#"))<=3){ /* #id */
        i=yytext+level+1;
        strcpy(buf,i+strspn(i,"_,^~*")); /* skip *bold* */
        *(buf+strcspn(buf,"\t "))='\0';
        printf("<p id=\"%s\">\n\n",buf);yyless(strspn(yytext,"\n")+1);
    } else {
        printf("<p>\n\n");yyless(strspn(yytext,"\n"));
    } yy_push_state(p);
}
<body>[\n]{0,2}("class"|"style"|"id")"=\""[^\"]+"\"" {
    printf("<p %s>\n\n",yytext);yy_push_state(p);
}
<p>{
\n\n {
    printf("</p>");
    yy_pop_state();
    }
}

    /* email address */
<p,bq,h,bo,it,table,div,xlist,sup,sub,strike,body>{
    [[:alnum:]_\-.]+"@"[[:alnum:]_\-.]+[[:alpha:]] {
        xurl=strstr(yytext,"@");*xurl++=0;
        printf("<a href=\"mailto:%s&#64;%s\">%s&#64;%s</a>",
            yytext,xurl,yytext,xurl);
        }
        /* Time machine */
        /* link 1-3 words matching html file names in current dir */
    ([^[:alnum:]_][[:alnum:]_]+){1,3}[^[:alnum:]_] {
        char *label, endchar=*(yytext+yyleng-1);
        LinkedList found;
        strncpy(buf,yytext+1,yyleng-1); buf[yyleng-2]='\0';
        label=strip(buf); // extra space and \n between words
        //~ fprintf(stderr,"nf? >%s<\n",label);
        if((found=target_exists(label))){
            xurl = url_encode(found->data);
            printf("%c<a href=\"%s\">%s</a>",*yytext,xurl,label);
            FREE(xurl);unput(endchar);
        } else {
            REJECT;
        }
    }
        /* link //... URLs */
    ("ftp"|"http"s?|"gopher"|"telnet":)?"//"[[:alnum:]\?&;=%#_\-./]+   {
            //~ char *href = url_encode(yytext);strspn
            printf("<a href=\"%s\" target=\"_blank\">%s</a>",yytext,strstr(yytext,"//")+2);
        }
        /* link [http://... text or image] */
    "["[^\[\]]+"]" {
            yytext[yyleng-1]='\0'; //erase beginning and end chars '[]'
            yytext++;
            sscanf(yytext,"%ms",&xurl); //get URL
            //~ no spaces?
            if(!strchr(yytext,' ')){
                //~ [IMAGE] ? [IMAGE] : [URL]
                if(is_image(xurl)){
                    printf("<img src=\"%s\" alt=\"%s\">", yytext, yytext);
                } else {
                    printf("<a href=\"%s\">%s</a>",yytext, yytext);
                }
            //~ spaces then
            } else {
                //~ clip off last portion
                char *last=strrchr(yytext, ' ');*(last++)='\0';
                //~ is last portion an image?
                if(is_image(last)){
                    //~ is first portion an attribute?
                    if(strchr(xurl,'"') || strchr(xurl,'\'')){ // [ATTRS IMAGE]
                        printf("<img %s src=\"%s\">",yytext, last);
                    //~ not an attribute then, must be URL
                    } else { // [URL ATTRS IMAGE]
                        if(strchr(yytext,' ')) yytext=strchr(yytext,' ')+1;
                        printf("<a href=\"%s\"><img %s src=\"%s\"></a>",xurl, yytext, last);
                    }
                //~ no image and has spaces
                } else {
                    //~ Assume [URL ATTRS DESC]
                    if(strchr(yytext,' ')){ // has spaces? (possible attributes)
                        printf("<a href=\"%s\"",xurl);
                        yytext=strchr(yytext,' ')+1;FREE(tmp);
                        //~ print any attributes
                        while(strstr(yytext,"=\"")||strstr(yytext,"= \"")) {
                            tmp=yytext;
                            //~ move scanner to end of attribute
                            yytext=strchr(yytext,'"')+1;
                            yytext=strchr(yytext,'"')+1;
                            *(yytext++)='\0'; //tmp is now print-ready
                            printf("%s ",tmp);
                        } // cose tag and print the description
                        printf(">%s %s</a>",yytext,last);
                    } else { // no spaces [URL DESC]
                        printf("<a href=\"%s\">%s</a>",xurl, last);
                    }
                }
            }
            FREE(xurl); FREE(tmp);
        }
}

    /* blockquote (indented text) */
<body,bq>^[\t ]+[^\t#*< ] {BEGIN(bq);
    indent_level = countIndent(yytext);
    if(indent_level > indents){
        for(int i=indent_level;i>indents;i--)
            printf("<blockquote>\n");
        indents=indent_level;
    }
    if(indent_level < indents){
        for(int i=indent_level;i < indents;i++)
            printf(" </blockquote>");
        indents=indent_level;
    }
    yyless(1);
}
<bq>{
[\n]{2,}  {
        for(;indent_level--;){
            printf(" </blockquote>");
        }
        indents=0;indent_level=0;
        BEGIN(body);
    }
}

    /* list */
^[\t ]+("*"|"#")[\t ] {
    indent_level = countIndent(yytext);
    if(indent_level > indents){
        for(int i=indent_level;i>indents;i--){
            if(yytext[yyleng-2]=='*') {
                printf("<ul>\n");tagList=list(tagList,PREPEND,"</ul>\n");
            } else {
                printf("<ol>\n");tagList=list(tagList,PREPEND,"</ol>\n");
            }
        }
        indents=indent_level;
    }
    if(indent_level < indents){
        for(int i=indent_level;i < indents;i++) {
            printf("%s",(char*)list(tagList,POP,0));
        indents=indent_level;
        }
    }printf("<li>");BEGIN(xlist);
}
<xlist>{
.$ {
        printf("%s </li>",yytext);
    }
\n\n {
        for(;indent_level--;){
            if(tagList)printf("%s",(char*)list(tagList,POP,indent_level));
        }
        indents=0;indent_level=0;        
        BEGIN(body); //tagList=list(tagList,DESTROY);
    }
}

    /* italics */
"_".*"_"  {
        yyless(1);
        printf("<i>");yy_push_state(it);
    }
<it>{
."_"     printf("%c</i>",yytext[0]);yy_pop_state();
"_"     printf("</i>");yy_pop_state();
}

    /* bold */
"*".*"*" {
        yyless(1);
        printf("<b>"); yy_push_state(bo);
    }
<bo>{
."*"     printf("%c</b>",yytext[0]);yy_pop_state();
}

    /* sup */
"^".*"^" {
        yyless(1);
        printf("<sup>"); yy_push_state(sup);
    }
<sup>{
."^"    printf("%c</sup>",yytext[0]);yy_pop_state();
}

    /* sub */
",,".*",," {yyless(2);
        printf("<sub>"); yy_push_state(sub);
    }
<sub>{
",,"    printf("</sub>");yy_pop_state();
}

    /* strike */
"~~".*"~~" {yyless(2);
        printf("<strike>"); yy_push_state(strike);
        }
<strike>{
"~~"    printf("</strike>");yy_pop_state();
}


    /* code class */
"{{{"[\t ]*("class="|"style=").*$ {
        int tabs = strspn(yytext+3,"\t ");
        printf("\n\n<code %s>",yytext+3+tabs);
        yy_push_state(code);header_sent=2;
        }
    /* code */
"{{{"[\n]* { /* eat returns */
        printf("\n\n<code class=\"prettyprint\">");
        yy_push_state(code);header_sent=2;
        }
<code>{
\n      { /* skip first return */
        if (header_sent==2) printf("<br/>\n");
        else header_sent = 2;
        }
"<"     printf("&lt;");
">"     printf("&gt;");
"&"     printf("&amp;");
" "    printf("&nbsp;");
\n*"}}}"\n* {
        printf("</code>");
        header_sent=1;
        yy_pop_state();
    }
}

    /* span */
"`"[\n]* { /* eat returns */
        printf("<span>"); yy_push_state(span);
    }
<span>{
"<"     printf("&lt;");
">"     printf("&gt;");
"&"     printf("&amp;");
" "    printf("&nbsp;");
\n*"`" {
        printf("</span>");
        yy_pop_state();
    }
}

     /* include file */
<*>^".include"[\t ]+  {yy_push_state(inc);
}
<inc>[[:alnum:]/\-._ ]+$ {
    /* if extension is .tml, process it first */
    char *ext=strrchr(yytext,'.');
    FILE *f;
    if(ext&&!strcmp(ext,".tml")){
        strcpy(buf,"tml -qr "); strncat(buf,yytext,1010);
        f=popen(buf,"r");
        if(f){
            while (fgets(buf,1024,f)!=NULL)
                printf("%s",buf);
            pclose(f);
        } else perror(".include tml file");
    }
    else { /* otherwise include it directly */
        f=fopen(yytext,"r");
        if(f){
            while (fgets(buf,1024,f)!=NULL)
                printf("%s",buf);
            fclose(f);
        }
    }
    yy_pop_state();
}

    /* table */
<body>^[^|\n{`]+"||" {
        strcpy(buf,yytext);
        buf[yyleng-2]='\0'; /* strip off || */
        printf("<table summary=\"%s\">\n<tr><th>",strip(buf));yyless(0);
        yy_push_state(table);
    }
<table>{
\n[^\n]  { /* end line */
        strcpy(buf,yytext);
        /* first line is table header <th> */
        if(header_sent==1)  {
            printf("</th></tr>\n<tr><td>%s",strip(buf));
            header_sent=2; /* subsequent lines are <td> */
        }
        else {
            printf("</td></tr>\n<tr><td>%s",strip(buf));
        }
    }
"||" { /* table first line is table header <th> */
        if(header_sent==1) printf("</th><th>");
        else printf("</td><td>");
    }
\n\n  { /* end table */
        printf("</td></tr>\n</table>");
        header_sent=1;
        yy_pop_state(); 
    }
}

    /* image button */
<p,bq,h,bo,it,table,div,xlist,sup,sub,strike,body>"[["    {
      printf("<a");yy_push_state(but);  
}
<but>{
    /* "title="[^ ]+ ECHO; */
[[:alnum:]_\-.]+"@"[[:alnum:]_\-.]+ {
    xurl=strstr(yytext,"@");*xurl++=0;
    printf("href=\"mailto:%s&#64;%s\">\n%s</a>",
        yytext,xurl,buf);
    yy_pop_state();
    }
[^@ \n]+$ { /* match url */
        printf("href=\"%s\">\n%s</a>",yytext,buf);
        yy_pop_state();
    }
[^\]\n]+"]]" { /* copy [button text] into buffer */
        strcpy(buf,yytext);
        buf[yyleng-2]=0;
    }
}

    /* pass raw contents of tags, especially php and comments */
"<?"(.|\n)*"?>"  ECHO;
"<!--"(.|\n)*"--"[\t ]?">"  ECHO;
"<"[^>]+">"  ECHO;
"<script"[^\n]+"</script>"  ECHO;

    /* pass htmlentities */
"&"[[:alnum:]#]+";"    ECHO;

    /* HTML entities */

<p,bq,h,bo,it,table,div,xlist,sup,sub,strike>{
    "\""    printf("&quot;");
    "&"     printf("&amp;");
    "™"|"(TM)"  printf("<sup>&trade;</sup>");
    "®"|"(R)"   printf("<sup>&reg;</sup>");
    "©"|"(C)"   printf("&copy;");
    "¶"|"(PP)"  printf("&para;");
    "℗"|"(P)"   printf("&pub;");
    "¢"|"(cents)"   printf("&cent; ");
}

<body><<EOF>>         {
    BEGIN(INITIAL);
    
    /* garbage collection */
    list(fileList,DESTROY);
    list(tagList,DESTROY);
#if !YY_FLEX_SUBMINOR_VERSION || YY_FLEX_SUBMINOR_VERSION < 27
// The versions of flex before 3.5.27 do not free their stack when done, so explcitly free it.
// Note that versions before did not actually define a subminor macro.
if (yy_start_stack) {
    yy_flex_free(yy_start_stack);
    yy_start_stack = 0;
}
#endif
#if YY_FLEX_SUBMINOR_VERSION >= 31 && YY_FLEX_SUBMINOR_VERSION < 34
// The versions of flex starting 3.5.31 do not free yy_buffer_stack, so do so
// explicitly (first yy_delete_buffer must be called to free pointers stored on the stack, then it is
// safe to remove the stack). This was fixed in 3.4.34.
    if(yy_buffer_stack) {
        yy_delete_buffer(YY_CURRENT_BUFFER);
        yyfree(yy_buffer_stack);
        yy_buffer_stack = 0;
    }
#endif
    yyterminate();
}

%%
LinkedList test_ext(char *name, char *ext, char *val) {
    strcpy(ext,val);
    return list(fileList,_FIND_PTR,name,strcasecmp);
}
/* test if word matches any html file names in list './' */
/* FREE the result! */
LinkedList target_exists(const char *in_p) {
    debug("testing [%s]\n",in_p);
    LinkedList lst=NULL;
    char *ext,*name,*tests[12]={
    "",
    "s",
    ".html",
    "s.html",
    ".htm",
    "s.htm",
    ".php",
    "s.php",
    ".txt",
    "s.txt",
    ".asp",
    "s.asp"
    };
    TRY(name=malloc(strlen(in_p)+10));
    strcpy(name,in_p);
    ext=name+strlen(name);
    for(int i=0;i<12;i++){
        lst=test_ext(name,ext,tests[i]);
        if(lst) break;
    }
    free(name);
    return lst;
}

int is_html(char *name){
    char *ext;
    ext = strrchr(name,'.');
    if(ext&&(
        !strcmp(ext,".html")||
        !strcmp(ext,".htm")||
        !strcmp(ext,".php")||
        !strcmp(ext,".txt")||
        !strcmp(ext,".asp")))
            return -1;
    else return 0;
}

int is_image(char *buf){
    char *ext;
    ext = strrchr(buf,'.');
    if(ext)
        if(!strcmp(ext,".jpg")||!strcmp(ext,".gif")||!strcmp(ext,".png")) return -1;
    return 0;
}

int read_dir(char *fname){ /* read directory entries */
	DIR *dp;
    int len;
	struct dirent *ep;
	dp = opendir("./");
	if (dp != NULL) {
		while ((ep = readdir(dp))) /* get all .html files */
            if(is_html(ep->d_name)) {
                /* except ourselves */
                len = strrchr(ep->d_name,'.') - ep->d_name;
                if(ep->d_name&&!strncmp(ep->d_name,fname,len)) continue;
                /* add to file list */
                fileList = list(fileList,PREPEND,ep->d_name);
            }
		(void)closedir(dp);
        //~ list(fileList,PRINT,1);
	} else
		perror("Couldn't open working directory");
	return 0;
}

int countIndent(char *text){
    int i=0,spaces=0,tabs=0,label=1;
    char *s=text;
    while(s[i]&&(s[i]==' '||s[i]=='\t'||label)){
        if(s[i]==' '||s[i]=='\t')label=0;
        else spaces++;
        if(s[i]=='\t')tabs++;
        if(s[i]==' ')spaces++;
        i++;
    }
    return spaces/4+tabs;
}

void welcome(void){
    fprintf (stderr,"TML (C) 2013 Henry Kroll www.anch.org\n");
}

void help(void){ printf(
"Usage: tml [options] file\n"
"-quiet | -q  Suppress welcome message.\n"
"-raw   | -r  Create raw HTML (skip header, body tags).\n\n");
}

int main(int argc, char **argv ){
    char *fname=NULL;
#define HAS(a) strstr(argv[argc],#a)
    if (argc){ /* parse command line options */
        for (--argc;argc;argc--){
            if (argv[argc][0]=='-'){
                if (HAS(r)) raw = 1;
                if (HAS(q)) quiet = 1;
                if (!(raw|quiet)) {
                    help();
                    return 0;
                }
            } else fname = argv[argc];
        }
        if (!fname)goto instd;
        if (!quiet) welcome();
        read_dir(fname);
        yyin = fopen(fname,"r");
        if (!yyin) { perror("fopen");
            exit (1);
        }
    } else {
instd:  if(!quiet){ welcome();
            fprintf (stderr,
            "No file specified. Reading stdin... Press CTRL-C to cancel.\n");
        }
        yyin = stdin;
    }
    if(!raw){
        /* print default HTML header */
        printf("<!DOCTYPE html><html><head>\n");
    }
    yylex();
    if(!raw) printf ("</body></html>\n");
    return 0;
    /* never reached: suppress warnings */
    yyunput(0,NULL);input();
}
